/*
  We want to demonstrate that the "condition variable queue"
  is not the same as the "mutex queue".
*/
#include <windows.h>   // Win32 Sleep()
#include <unistd.h>    // Linux sleep()
#include <stdio.h>
#include <pthread.h>

pthread_mutex_t cv_mutex;
pthread_cond_t  my_condition_variable;

int main()
{
   //thread function prototypes
   void *thread1(void *vargp);
   void *thread2(void *vargp);
   void *thread3(void *vargp);
   void *thread4(void *vargp);
   void *thread5(void *vargp);

   pthread_t tid[5];

   // initialize mutex and condition variable objects
   pthread_mutex_init(&cv_mutex, NULL);
   pthread_cond_init (&my_condition_variable, NULL);

   // create the three threads
   printf("Starting threads.\n");
   pthread_create(&(tid[0]), NULL, thread1, NULL);  // should have error checking
   pthread_create(&(tid[1]), NULL, thread2, NULL);
   pthread_create(&(tid[2]), NULL, thread3, NULL);
   pthread_create(&(tid[3]), NULL, thread4, NULL);
   pthread_create(&(tid[4]), NULL, thread5, NULL);

   // wait for each thread to complete
   pthread_join(tid[0], NULL);
   pthread_join(tid[1], NULL);
   pthread_join(tid[2], NULL);
   pthread_join(tid[3], NULL);
   pthread_join(tid[4], NULL);

   // clean up and exit
   printf("Cleaning up threads.\n");
   pthread_mutex_destroy(&cv_mutex);
   pthread_cond_destroy(&my_condition_variable);
   pthread_exit(NULL);
   return 0;  // keep gcc happy
}//main


void *thread1(void *vargp)
{
   printf("Thread 1 grabs the mutex.\n");
   pthread_mutex_lock(&cv_mutex);
   printf("Thread 1 blocks on the CV. (And gives up the mutex.)\n");
   pthread_cond_wait(&my_condition_variable, &cv_mutex);
   printf("Thread 1 released from the CV! (And reaquires the mutex.)\n");
   printf("Thread 1 gives up the mutex.\n");
   pthread_mutex_unlock(&cv_mutex);

   pthread_exit(NULL);
   return NULL;  // keep gcc happy
}//thread1


void *thread2(void *vargp)
{
   Sleep(4000);
   printf("Thread 2 grabs the mutex.\n");
   pthread_mutex_lock(&cv_mutex);
   Sleep(5000);
   printf("Thread 2 gives up the mutex.\n");
   pthread_mutex_unlock(&cv_mutex);

   pthread_exit(NULL);
   return NULL;  // keep gcc happy
}//thread2


void *thread3(void *vargp)
{
   Sleep(4000);
   printf("Thread 3 grabs the mutex.\n");
   pthread_mutex_lock(&cv_mutex);
   Sleep(5000);
   printf("Thread 3 gives up the mutex.\n");
   pthread_mutex_unlock(&cv_mutex);

   pthread_exit(NULL);
   return NULL;  // keep gcc happy
}//thread3


void *thread4(void *vargp)
{
   Sleep(4000);
   printf("Thread 4 grabs the mutex.\n");
   pthread_mutex_lock(&cv_mutex);
   Sleep(5000);
   printf("Thread 4 gives up the mutex.\n");
   pthread_mutex_unlock(&cv_mutex);  // Try commenting this out. What happens?

   pthread_exit(NULL);
   return NULL;  // keep gcc happy
}//thread4


void *thread5(void *vargp)
{
   Sleep(6000);
   printf("Thread 5 signals the CV.\n");
   pthread_cond_signal(&my_condition_variable);

   pthread_exit(NULL);
   return NULL;  // keep gcc happy
}//thread5