/*
   Page 12 of
   http://pages.cs.wisc.edu/~remzi/OSTEP/threads-cv.pdf

   Figure 30.11: The Correct Put And Get Routines
*/
#include <stdio.h>
#include "mythreads.h"

#define MAX 40  /*** author forgot this ***/
//#define MAX 39  /* This doesn't work! */

int buffer[MAX];
int fill_ptr = 0;
int use_ptr = 0;
int count = 0;

void put(int value)
{
   buffer[fill_ptr] = value;
   fill_ptr = (fill_ptr + 1) % MAX;
   count++;
}

int get()
{
   int tmp = buffer[use_ptr];
   use_ptr = (use_ptr + 1) % MAX;
   count--;
   return tmp;
}

/*
   Figure 30.12: The Correct Producer/Consumer Synchronization
*/

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t  empty = PTHREAD_COND_INITIALIZER;
pthread_cond_t   fill = PTHREAD_COND_INITIALIZER;

void *producer(void *arg)
{
   int i;
   int loops = (int) arg;   /*** author forgot this ***/
   for (i = 0; i < loops; i++)
   {
      Pthread_mutex_lock(&mutex);           // p1
      while (count == MAX)                  // p2
         pthread_cond_wait(&empty, &mutex); // p3
      put(i);                               // p4
      pthread_cond_signal(&fill);           // p5
      Pthread_mutex_unlock(&mutex);         // p6
   }
   return NULL; /*** author forgot this ***/
}

void *consumer(void *arg)
{
   int i;
   int loops = (int) arg;   /*** author forgot this ***/
   for (i = 0; i < loops; i++)
   {
      Pthread_mutex_lock(&mutex);          // c1
      while (count == 0)                   // c2
         pthread_cond_wait(&fill, &mutex); // c3
      int tmp = get();                     // c4
      pthread_cond_signal(&empty);         // c5
      Pthread_mutex_unlock(&mutex);        // c6
      printf("%d\n", tmp); /*** Why is this outside the critical section? ***/
   }
   return NULL; /*** author forgot this ***/
}


int
main( )
{
   pthread_t p1, p2, p3;
   printf("main: begin\n");
   Pthread_create(&p1, NULL, producer, (void*)40);
   Pthread_create(&p2, NULL, consumer, (void*)20);
   Pthread_create(&p3, NULL, consumer, (void*)20);
   // join waits for the producer to finish
   Pthread_join(p1, NULL);
   Pthread_join(p2, NULL);
   Pthread_join(p3, NULL);
   printf("main: end\n");
   return 0;
}
